// Copyright (C) 2010  Internet Systems Consortium, Inc. ("ISC")
//
// Permission to use, copy, modify, and/or distribute this software for any
// purpose with or without fee is hereby granted, provided that the above
// copyright notice and this permission notice appear in all copies.
//
// THE SOFTWARE IS PROVIDED "AS IS" AND ISC DISCLAIMS ALL WARRANTIES WITH
// REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS.  IN NO EVENT SHALL ISC BE LIABLE FOR ANY SPECIAL, DIRECT,
// INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
// LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
// OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
// PERFORMANCE OF THIS SOFTWARE.

// $Id$

#include <gtest/gtest.h>
#include <boost/shared_ptr.hpp>

#include <dns/rrclass.h>
#include <dns/rdataclass.h>

#include "../asiolink.h"
#include "../zone_entry.h"
#include "../nameserver_entry.h"
#include "../address_request_callback.h"
#include "../nsas_entry_compare.h"
#include "../hash_deleter.h"

#include "nsas_test.h"

using namespace isc::nsas;
using namespace asiolink;
using namespace std;
using namespace boost;
using namespace isc::dns;

namespace {

/// \brief Inherited version with access into its internals for tests
class InheritedZoneEntry : public ZoneEntry {
    public:
        InheritedZoneEntry(shared_ptr<ResolverInterface> resolver,
            const std::string& name, const RRClass& class_code,
            shared_ptr<HashTable<NameserverEntry> > nameserver_table,
            shared_ptr<LruList<NameserverEntry> > nameserver_lru) :
            ZoneEntry(resolver, name, class_code, nameserver_table,
                nameserver_lru)
        { }
        NameserverVector& nameservers() { return nameservers_; }
};

/// \brief Test Fixture Class
class ZoneEntryTest : public TestWithRdata {
protected:
    /// \brief Constructor
    ZoneEntryTest() :
        nameserver_table_(new HashTable<NameserverEntry>(
            new NsasEntryCompare<NameserverEntry>)),
        nameserver_lru_(new LruList<NameserverEntry>(
            (3 * nameserver_table_->tableSize()),
            new HashDeleter<NameserverEntry>(*nameserver_table_))),
        resolver_(new TestResolver),
        callback_(new Callback)
    { }
    /// \brief Tables of nameservers to pass into zone entry constructor
    shared_ptr<HashTable<NameserverEntry> > nameserver_table_;
    shared_ptr<LruList<NameserverEntry> > nameserver_lru_;
    /// \brief The resolver
    shared_ptr<TestResolver> resolver_;

    struct Callback : public AddressRequestCallback {
        Callback() : unreachable_count_(0) {}
        size_t unreachable_count_;
        vector<IOAddress> successes_;
        virtual void unreachable() { unreachable_count_ ++; }
        virtual void success(const IOAddress& address) {
            successes_.push_back(address);
        }
    };
    shared_ptr<Callback> callback_;

    // Empty callback to pass to nameserver entry, to do injection of them
    struct NseCallback : public NameserverEntry::Callback {
        virtual void operator()(shared_ptr<NameserverEntry>) { }
    };

    shared_ptr<NseCallback> nseCallback() {
        return (shared_ptr<NseCallback>(new NseCallback));
    }

    /**
     * \short Function returning a new zone.
     *
     * Convenience funcion, just creating a new zone, to shorten the code.
     */
    shared_ptr<InheritedZoneEntry> getZone() {
        return (shared_ptr<InheritedZoneEntry>(new InheritedZoneEntry(
            resolver_, EXAMPLE_CO_UK, RRClass::IN(), nameserver_table_,
            nameserver_lru_)));
    }

    /// \short Creates and injects a nameserver entry
    shared_ptr<NameserverEntry> injectEntry() {
        shared_ptr<NameserverEntry> nse(new NameserverEntry(ns_name_.toText(),
            RRClass::IN()));
        nameserver_table_->add(nse, HashKey(ns_name_.toText(), RRClass::IN()));
        EXPECT_EQ(Fetchable::NOT_ASKED, nse->getState());
        return (nse);
    }

    void checkInjected(bool answer, size_t success_count = 1,
        size_t failure_count = 0)
    {
        // Create the zone and check it acts correctly
        shared_ptr<InheritedZoneEntry> zone(getZone());
        resolver_->addPresetAnswer(Question(Name(EXAMPLE_CO_UK), RRClass::IN(),
            RRType::NS()), rr_single_);
        zone->addCallback(callback_, ANY_OK);
        EXPECT_EQ(2, resolver_->requests.size());
        EXPECT_TRUE(resolver_->asksIPs(ns_name_, 0, 1));
        if (answer) {
            EXPECT_NO_THROW(resolver_->answer(0, ns_name_, RRType::A(),
                rdata::in::A("192.0.2.1")));
            EXPECT_NO_THROW(resolver_->answer(1, ns_name_, RRType::AAAA(),
                rdata::in::AAAA("2001:db8::1")));
        }
        // Check the answers
        EXPECT_EQ(Fetchable::READY, zone->getState());
        EXPECT_EQ(failure_count, callback_->unreachable_count_);
        EXPECT_EQ(success_count, callback_->successes_.size());
        for (size_t i = 0; i < callback_->successes_.size(); ++ i) {
            EXPECT_TRUE(IOAddress("192.0.2.1").equal(
                callback_->successes_[i]) || IOAddress("2001:db8::1").equal(
                    callback_->successes_[i]));
        }
    }
};

/// Tests of the default constructor
TEST_F(ZoneEntryTest, DefaultConstructor) {

    // Default constructor should not create any RRsets
    InheritedZoneEntry alpha(resolver_, EXAMPLE_CO_UK,
        RRClass::IN(), nameserver_table_, nameserver_lru_);
    EXPECT_EQ(EXAMPLE_CO_UK, alpha.getName());
    EXPECT_EQ(RRClass::IN(), alpha.getClass());
    EXPECT_TRUE(alpha.nameservers().empty());
}

// It should answer negatively right away if there are no nameservers
TEST_F(ZoneEntryTest, CallbackNoNS) {
    shared_ptr<InheritedZoneEntry> zone(getZone());
    EXPECT_EQ(Fetchable::NOT_ASKED, zone->getState());
    // It should accept the callback
    zone->addCallback(callback_, ANY_OK);
    EXPECT_EQ(Fetchable::IN_PROGRESS, zone->getState());
    // Ask for the nameservers
    EXPECT_NO_THROW(resolver_->provideNS(0, rr_empty_));
    // And tell imediatelly that it is unreachable (when it has no nameservers)
    EXPECT_TRUE(callback_->successes_.empty());
    EXPECT_EQ(1, callback_->unreachable_count_);
}

/*
 * This checks that:
 * * The question is answered even with 0 TTL
 * * It is answered only once, other query triggers another upstream query.
 * * When we answer with different one, it should ask for a different zone.
 */
TEST_F(ZoneEntryTest, ChangedNS) {
    // Make it zero TTL, so it expires right away
    rr_single_->setTTL(RRTTL(0));
    shared_ptr<InheritedZoneEntry> zone(getZone());
    EXPECT_EQ(Fetchable::NOT_ASKED, zone->getState());
    // It should accept the callback
    zone->addCallback(callback_, ANY_OK);
    EXPECT_EQ(Fetchable::IN_PROGRESS, zone->getState());
    EXPECT_NO_THROW(resolver_->provideNS(0, rr_single_));
    // It should not be answered yet, it should ask for the IP addresses
    EXPECT_TRUE(callback_->successes_.empty());
    EXPECT_EQ(0, callback_->unreachable_count_);
    EXPECT_TRUE(resolver_->asksIPs(ns_name_, 1, 2));
    EXPECT_NO_THROW(resolver_->answer(1, ns_name_, RRType::A(),
        rdata::in::A("192.0.2.1")));
    ASSERT_EQ(1, callback_->successes_.size());
    EXPECT_TRUE(IOAddress("192.0.2.1").equal(callback_->successes_[0]));
    EXPECT_NO_THROW(resolver_->answer(2, ns_name_, RRType::AAAA(),
        rdata::in::AAAA("2001:db8::1")));
    EXPECT_EQ(1, callback_->successes_.size());
    // It should request the NSs again, as TTL is 0
    zone->addCallback(callback_, ANY_OK);
    EXPECT_EQ(4, resolver_->requests.size());
    EXPECT_EQ(Fetchable::IN_PROGRESS, zone->getState());

    Name different_name("ns.example.com");
    // Create a different answer
    RRsetPtr different_ns(new RRset(Name(EXAMPLE_CO_UK), RRClass::IN(),
        RRType::NS(), RRTTL(0)));
    different_ns->addRdata(rdata::generic::NS(different_name));
    EXPECT_NO_THROW(resolver_->provideNS(3, different_ns));
    // It should become ready and ask for the new name
    EXPECT_EQ(Fetchable::READY, zone->getState());
    // Answer one of the IP addresses, we should get an address now
    EXPECT_TRUE(resolver_->asksIPs(different_name, 4, 5));
    EXPECT_NO_THROW(resolver_->answer(4, different_name, RRType::A(),
        rdata::in::A("192.0.2.2")));
    ASSERT_EQ(2, callback_->successes_.size());
    EXPECT_TRUE(IOAddress("192.0.2.2").equal(callback_->successes_[1]));

    // And now, switch back, as it timed out again
    zone->addCallback(callback_, ANY_OK);
    EXPECT_EQ(7, resolver_->requests.size());
    EXPECT_EQ(Fetchable::IN_PROGRESS, zone->getState());
    // When we answer with the original, it should still be cached and
    // we should get the answer
    EXPECT_NO_THROW(resolver_->provideNS(0, rr_single_));
    EXPECT_EQ(7, resolver_->requests.size());
    EXPECT_EQ(Fetchable::READY, zone->getState());
    ASSERT_EQ(3, callback_->successes_.size());
    EXPECT_TRUE(IOAddress("192.0.2.1").equal(callback_->successes_[0]));
}

// Check it answers callbacks when we give it addresses
TEST_F(ZoneEntryTest, CallbacksAnswered) {
    shared_ptr<InheritedZoneEntry> zone(getZone());
    // It should be in NOT_ASKED state
    EXPECT_EQ(Fetchable::NOT_ASKED, zone->getState());
    // It should accept the callback
    zone->addCallback(callback_, ANY_OK);
    EXPECT_EQ(Fetchable::IN_PROGRESS, zone->getState());
    EXPECT_NO_THROW(resolver_->provideNS(0, rr_single_));
    // It should not be answered yet, it should ask for the IP addresses
    EXPECT_TRUE(callback_->successes_.empty());
    EXPECT_EQ(0, callback_->unreachable_count_);
    EXPECT_TRUE(resolver_->asksIPs(ns_name_, 1, 2));
    // We should be READY, as it marks we have nameservers
    // (not that they are ready)
    EXPECT_EQ(Fetchable::READY, zone->getState());
    // Give two more callbacks, with different address families
    zone->addCallback(callback_, V4_ONLY);
    zone->addCallback(callback_, V6_ONLY);
    // Nothing more is asked
    EXPECT_EQ(3, resolver_->requests.size());
    EXPECT_NO_THROW(resolver_->answer(1, ns_name_, RRType::A(),
         rdata::in::A("192.0.2.1")));
    // Two are answered (ANY and V4)
    ASSERT_EQ(2, callback_->successes_.size());
    EXPECT_TRUE(IOAddress("192.0.2.1").equal(callback_->successes_[0]));
    EXPECT_TRUE(IOAddress("192.0.2.1").equal(callback_->successes_[1]));
    // None are rejected
    EXPECT_EQ(0, callback_->unreachable_count_);
    EXPECT_NO_THROW(resolver_->answer(2, ns_name_, RRType::AAAA(),
        rdata::in::AAAA("2001:db8::1")));
    // This should answer the third callback
    EXPECT_EQ(0, callback_->unreachable_count_);
    ASSERT_EQ(3, callback_->successes_.size());
    EXPECT_TRUE(IOAddress("2001:db8::1").equal(callback_->successes_[2]));
    // It should think it is ready
    EXPECT_EQ(Fetchable::READY, zone->getState());
    // When we ask something more, it should be answered right away
    zone->addCallback(callback_, V4_ONLY);
    EXPECT_EQ(3, resolver_->requests.size());
    ASSERT_EQ(4, callback_->successes_.size());
    EXPECT_TRUE(IOAddress("192.0.2.1").equal(callback_->successes_[3]));
    EXPECT_EQ(0, callback_->unreachable_count_);
}

// Pretend the server can be reached only by IPv4
TEST_F(ZoneEntryTest, CallbacksAOnly) {
    shared_ptr<InheritedZoneEntry> zone(getZone());
    // It should be in NOT_ASKED state
    EXPECT_EQ(Fetchable::NOT_ASKED, zone->getState());
    // It should accept the callback
    zone->addCallback(callback_, ANY_OK);
    EXPECT_EQ(Fetchable::IN_PROGRESS, zone->getState());
    EXPECT_NO_THROW(resolver_->provideNS(0, rr_single_));
    // It should not be answered yet, it should ask for the IP addresses
    EXPECT_TRUE(callback_->successes_.empty());
    EXPECT_EQ(0, callback_->unreachable_count_);
    EXPECT_TRUE(resolver_->asksIPs(ns_name_, 1, 2));
    EXPECT_EQ(Fetchable::READY, zone->getState());
    // Give two more callbacks, with different address families
    zone->addCallback(callback_, V4_ONLY);
    zone->addCallback(callback_, V6_ONLY);
    ASSERT_GE(resolver_->requests.size(), 3);
    resolver_->requests[2].second->failure();
    // One should be rejected, but two still stay, they have chance
    EXPECT_EQ(0, callback_->successes_.size());
    EXPECT_EQ(1, callback_->unreachable_count_);
    // Answer the A one and see it answers what can be answered
    EXPECT_NO_THROW(resolver_->answer(1, ns_name_, RRType::A(),
        rdata::in::A("192.0.2.1")));
    ASSERT_EQ(2, callback_->successes_.size());
    EXPECT_TRUE(IOAddress("192.0.2.1").equal(callback_->successes_[0]));
    EXPECT_TRUE(IOAddress("192.0.2.1").equal(callback_->successes_[1]));
    EXPECT_EQ(1, callback_->unreachable_count_);
    // Everything arriwed, so we are ready
    EXPECT_EQ(Fetchable::READY, zone->getState());
    // Try asking something more
    zone->addCallback(callback_, V4_ONLY);
    EXPECT_EQ(3, resolver_->requests.size());
    ASSERT_EQ(3, callback_->successes_.size());
    EXPECT_TRUE(IOAddress("192.0.2.1").equal(callback_->successes_[2]));
    EXPECT_EQ(1, callback_->unreachable_count_);

    zone->addCallback(callback_, V6_ONLY);
    EXPECT_EQ(3, resolver_->requests.size());
    EXPECT_EQ(3, callback_->successes_.size());
    EXPECT_EQ(2, callback_->unreachable_count_);
}

// See it tries hard enough to get address and tries both nameservers
TEST_F(ZoneEntryTest, CallbackTwoNS) {
    shared_ptr<InheritedZoneEntry> zone(getZone());
    // It should be in NOT_ASKED state
    EXPECT_EQ(Fetchable::NOT_ASKED, zone->getState());
    // It should accept the callback
    zone->addCallback(callback_, V4_ONLY);
    EXPECT_EQ(Fetchable::IN_PROGRESS, zone->getState());
    EXPECT_NO_THROW(resolver_->provideNS(0, rrns_));
    EXPECT_EQ(Fetchable::READY, zone->getState());
    // It asks a question (we do not know which nameserver)
    shared_ptr<Name> name;
    ASSERT_NO_THROW(name.reset(new Name((*resolver_)[1]->getName())));
    ASSERT_TRUE(resolver_->asksIPs(*name, 1, 2));
    resolver_->requests[1].second->failure();
    // Nothing should be answered or failed yet
    EXPECT_EQ(0, callback_->unreachable_count_);
    EXPECT_EQ(0, callback_->successes_.size());
    ASSERT_TRUE(resolver_->asksIPs((*resolver_)[3]->getName(), 3, 4));
    // Fail the second one
    resolver_->requests[3].second->failure();
    // The callback should be failed now, as there is no chance of getting
    // v4 address
    EXPECT_EQ(1, callback_->unreachable_count_);
    EXPECT_EQ(0, callback_->successes_.size());
    // And question for v6 or any should still wait while v4 should be failed
    // right away
    zone->addCallback(callback_, V6_ONLY);
    EXPECT_EQ(1, callback_->unreachable_count_);
    EXPECT_EQ(0, callback_->successes_.size());

    zone->addCallback(callback_, ANY_OK);
    EXPECT_EQ(1, callback_->unreachable_count_);
    EXPECT_EQ(0, callback_->successes_.size());

    zone->addCallback(callback_, V4_ONLY);
    EXPECT_EQ(2, callback_->unreachable_count_);
    EXPECT_EQ(0, callback_->successes_.size());
    // Answer the IPv6 one
    EXPECT_NO_THROW(resolver_->answer(2, (*resolver_)[2]->getName(),
        RRType::AAAA(), rdata::in::AAAA("2001:db8::1")));

    // Ready, as we have at last some address
    EXPECT_EQ(Fetchable::READY, zone->getState());
    // The other callbacks should be answered now
    EXPECT_EQ(2, callback_->unreachable_count_);
    ASSERT_EQ(2, callback_->successes_.size());
    EXPECT_TRUE(IOAddress("2001:db8::1").equal(callback_->successes_[0]));
    EXPECT_TRUE(IOAddress("2001:db8::1").equal(callback_->successes_[1]));
}

// Test if it works when the response comes right away from the resolve call
TEST_F(ZoneEntryTest, DirectAnswer) {
    shared_ptr<InheritedZoneEntry> zone(getZone());
    EXPECT_EQ(Fetchable::NOT_ASKED, zone->getState());

    // One unsuccessfull attempt, nameservers fail
    resolver_->addPresetAnswer(Question(Name(EXAMPLE_CO_UK), RRClass::IN(),
        RRType::NS()), shared_ptr<AbstractRRset>());
    zone->addCallback(callback_, ANY_OK);
    EXPECT_EQ(0, callback_->successes_.size());
    EXPECT_EQ(1, callback_->unreachable_count_);
    EXPECT_EQ(0, resolver_->requests.size());
    EXPECT_EQ(Fetchable::UNREACHABLE, zone->getState());

    // Successfull attempt now
    zone = getZone();
    // First, fill the answers to all the questions it should ask
    EXPECT_EQ(Fetchable::NOT_ASKED, zone->getState());
    resolver_->addPresetAnswer(Question(Name(EXAMPLE_CO_UK), RRClass::IN(),
        RRType::NS()), rr_single_);
    Name ns_name("ns.example.net");
    rrv4_->setName(ns_name);
    resolver_->addPresetAnswer(Question(ns_name, RRClass::IN(), RRType::A()),
        rrv4_);
    rrv6_->setName(ns_name);
    resolver_->addPresetAnswer(Question(ns_name, RRClass::IN(),
        RRType::AAAA()), rrv6_);
    // Reset the results
    callback_->unreachable_count_ = 0;
    // Ask for the IP address
    zone->addCallback(callback_, ANY_OK);
    // It should be answered right away, positively
    EXPECT_EQ(1, callback_->successes_.size());
    EXPECT_EQ(0, callback_->unreachable_count_);
    EXPECT_EQ(0, resolver_->requests.size());
    EXPECT_EQ(Fetchable::READY, zone->getState());

    // Reset the results
    callback_->successes_.clear();
    // Now, pretend we do not have IP addresses
    resolver_->addPresetAnswer(Question(ns_name, RRClass::IN(), RRType::A()),
        shared_ptr<AbstractRRset>());
    resolver_->addPresetAnswer(Question(ns_name, RRClass::IN(),
        RRType::AAAA()), shared_ptr<AbstractRRset>());
    // Get another zone and ask it again. It should fail.
    // Clean the table first, though, so it does not find the old nameserver
    nameserver_table_->remove(HashKey(ns_name.toText(), RRClass::IN()));
    zone = getZone();
    EXPECT_EQ(Fetchable::NOT_ASKED, zone->getState());
    zone->addCallback(callback_, ANY_OK);
    EXPECT_EQ(0, callback_->successes_.size());
    EXPECT_EQ(1, callback_->unreachable_count_);
    EXPECT_EQ(0, resolver_->requests.size());
    // It should be ready, but have no IP addresses on the nameservers
    EXPECT_EQ(Fetchable::READY, zone->getState());
}

// Checks it asks only for addresses when the addresses time out, not NSs
TEST_F(ZoneEntryTest, AddressTimeout) {
    shared_ptr<InheritedZoneEntry> zone(getZone());
    // It should be in NOT_ASKED state
    EXPECT_EQ(Fetchable::NOT_ASKED, zone->getState());
    // It should accept the callback
    zone->addCallback(callback_, ANY_OK);
    EXPECT_EQ(Fetchable::IN_PROGRESS, zone->getState());
    EXPECT_NO_THROW(resolver_->provideNS(0, rr_single_));
    // It should not be answered yet, it should ask for the IP addresses
    EXPECT_TRUE(callback_->successes_.empty());
    EXPECT_EQ(0, callback_->unreachable_count_);
    EXPECT_TRUE(resolver_->asksIPs(ns_name_, 1, 2));
    // We should be READY, as it marks we have nameservers
    // (not that they are ready)
    EXPECT_EQ(Fetchable::READY, zone->getState());
    EXPECT_NO_THROW(resolver_->answer(1, ns_name_, RRType::A(),
         rdata::in::A("192.0.2.1"), 0));
    ASSERT_EQ(1, callback_->successes_.size());
    EXPECT_TRUE(IOAddress("192.0.2.1").equal(callback_->successes_[0]));
    // None are rejected
    EXPECT_EQ(0, callback_->unreachable_count_);
    EXPECT_NO_THROW(resolver_->answer(2, ns_name_, RRType::AAAA(),
        rdata::in::AAAA("2001:db8::1"), 0));
    EXPECT_EQ(1, callback_->successes_.size());
    EXPECT_EQ(Fetchable::READY, zone->getState());
    // When we ask for another one, it should ask for the addresses again
    zone->addCallback(callback_, ANY_OK);
    EXPECT_TRUE(resolver_->asksIPs(ns_name_, 3, 4));
    EXPECT_EQ(0, callback_->unreachable_count_);
    EXPECT_EQ(1, callback_->successes_.size());
    EXPECT_NO_THROW(resolver_->answer(3, ns_name_, RRType::A(),
         rdata::in::A("192.0.2.1"), 0));
    EXPECT_EQ(0, callback_->unreachable_count_);
    ASSERT_EQ(2, callback_->successes_.size());
    EXPECT_TRUE(IOAddress("192.0.2.1").equal(callback_->successes_[1]));
}

// Test how the zone reacts to a nameserver entry in ready state
TEST_F(ZoneEntryTest, NameserverEntryReady) {
    // Inject the entry
    shared_ptr<NameserverEntry> nse(injectEntry());
    // Fill it with data
    nse->askIP(resolver_, nseCallback(), ANY_OK);
    EXPECT_EQ(Fetchable::IN_PROGRESS, nse->getState());
    EXPECT_TRUE(resolver_->asksIPs(ns_name_, 0, 1));
    EXPECT_NO_THROW(resolver_->answer(0, ns_name_, RRType::A(),
        rdata::in::A("192.0.2.1")));
    EXPECT_NO_THROW(resolver_->answer(1, ns_name_, RRType::AAAA(),
        rdata::in::AAAA("2001:db8::1")));
    EXPECT_EQ(Fetchable::READY, nse->getState());

    checkInjected(false);
}

// Test how the zone reacts to a nameserver in not asked state
TEST_F(ZoneEntryTest, NameserverNotAsked) {
    // Inject the entry
    injectEntry();
    // We do not need it, nothing to modify on it

    checkInjected(true);
}

// What if the zone finds a nameserver in progress?
TEST_F(ZoneEntryTest, NameserverInProgress) {
    // Prepare the nameserver entry
    shared_ptr<NameserverEntry> nse(injectEntry());
    nse->askIP(resolver_, nseCallback(), ANY_OK);
    EXPECT_EQ(Fetchable::IN_PROGRESS, nse->getState());
    EXPECT_TRUE(resolver_->asksIPs(ns_name_, 0, 1));

    checkInjected(true);
}

// Check Zone's reaction to found expired nameserver
TEST_F(ZoneEntryTest, NameserverExpired) {
    shared_ptr<NameserverEntry> nse(injectEntry());
    nse->askIP(resolver_, nseCallback(), ANY_OK);
    EXPECT_EQ(Fetchable::IN_PROGRESS, nse->getState());
    EXPECT_TRUE(resolver_->asksIPs(ns_name_, 0, 1));
    EXPECT_NO_THROW(resolver_->answer(0, ns_name_, RRType::A(),
        rdata::in::A("192.0.2.1"), 0));
    EXPECT_NO_THROW(resolver_->answer(1, ns_name_, RRType::AAAA(),
        rdata::in::AAAA("2001:db8::1"), 0));
    EXPECT_EQ(Fetchable::READY, nse->getState());
    NameserverEntry::AddressVector addresses;
    EXPECT_EQ(Fetchable::EXPIRED, nse->getAddresses(addresses));
    EXPECT_EQ(Fetchable::EXPIRED, nse->getState());
    resolver_->requests.clear();

    checkInjected(true);
}

// Check how it reacts to an unreachable zone already in the table
TEST_F(ZoneEntryTest, NameserverUnreachable) {
    shared_ptr<NameserverEntry> nse(injectEntry());
    nse->askIP(resolver_, nseCallback(), ANY_OK);
    ASSERT_EQ(2, resolver_->requests.size());
    resolver_->requests[0].second->failure();
    resolver_->requests[1].second->failure();
    EXPECT_EQ(Fetchable::UNREACHABLE, nse->getState());

    checkInjected(false, 0, 1);
}

}   // namespace
